import io
import json
import logging
from typing import Tuple

import avro
import avro.datafile
import avro.io
import avro.schema

from .codec import Codec

prefix = b"d_"


class AvroCodec(Codec):
    """
    Avro Codec

    """

    logger = logging.getLogger(__name__)
    schema_cache_max_n = 1000
    __blank_schema = {
        "type": "record",
        "namespace": "streaminghub.datamux.data",
        "name": None,
        "fields": None,
    }
    __type_map = {
        bool: "boolean",
        int: "int",
        float: "float",
        bytes: "bytes",
        str: "string",
    }

    def __init__(self) -> None:
        self.__schema_registry = {}

    def __create_schema(
        self,
        subtopic: bytes,
        content: dict,
    ) -> dict:
        schemadict = self.__blank_schema.copy()
        schemadict["name"] = subtopic.decode()
        schemadict["fields"] = []
        for k, v in content.items():
            schemadict["fields"].append(
                {
                    "name": k,
                    "type": self.__type_map[type(v)],
                    "doc": "...",
                }
            )
        return schemadict

    def __preprocess(
        self,
        content: dict,
    ) -> dict:
        flat_dict = {}
        for root_k in content.keys():
            for inner_k, v in content[root_k].items():
                flat_dict[f"{root_k}.{inner_k}"] = v
        return flat_dict

    def __postprocess(
        self,
        content: dict,
    ) -> dict:
        nested_dict = {}
        for flat_k, v in content.items():
            root_k, inner_k = flat_k.split(".", 1)
            if root_k not in nested_dict:
                nested_dict[root_k] = {}
            nested_dict[root_k][inner_k] = v
        return nested_dict

    def __cache_schema(
        self,
        subtopic: bytes,
        schema: avro.schema.Schema,
    ) -> None:
        self.__schema_registry[subtopic] = schema
        # silently prune cache
        N = self.schema_cache_max_n
        for key in list(self.__schema_registry)[:-N]:
            self.__schema_registry.pop(key, None)

    def __encode_avro(
        self,
        subtopic: bytes,
        content: dict,
    ) -> bytes:
        # get schema
        schema = self.__schema_registry[subtopic]
        writer = avro.io.DatumWriter(schema)
        # encoding
        buffer = io.BytesIO()
        encoder = avro.io.BinaryEncoder(buffer)
        writer.write(content, encoder)
        content_bytes = buffer.getvalue()
        # return encoded data
        return content_bytes

    def __decode_avro(
        self,
        content_bytes: bytes,
        schema: avro.schema.Schema,
    ) -> dict:
        # create reader
        reader = avro.io.DatumReader(schema)
        # decoding
        buffer = io.BytesIO(content_bytes)
        decoder = avro.io.BinaryDecoder(buffer)
        content = reader.read(decoder)
        # return decoded data
        return dict(content)  # type: ignore

    def __encode_json(
        self,
        content: dict,
    ) -> bytes:
        content_str = json.dumps(content)
        content_bytes = content_str.encode()
        return content_bytes

    def __decode_json(
        self,
        content_bytes: bytes,
    ) -> dict:
        content_str = content_bytes.decode()
        content = json.loads(content_str)
        return content

    def encode(
        self,
        topic: bytes,
        content: dict,
    ) -> list[bytes]:
        self.logger.debug(f"encode(): topic={topic}, content={content}")
        lines = []
        if len(content) == 0:
            self.logger.info(f"encode(): got empty message - topic={topic}")
            content_enc = b""
            lines.append(topic + b"||" + content_enc)
        elif topic.startswith(prefix) and len(content) > 0:
            # preprocessing
            subtopic = topic[len(prefix) :]
            content = self.__preprocess(content)
            # ensure schema is registered
            if subtopic not in self.__schema_registry:
                # create schema
                schema = self.__create_schema(subtopic, content)
                # register schema in __schema_registry
                self.__cache_schema(subtopic, avro.schema.make_avsc_object(schema))
                self.logger.debug(f"encode(): assigned schema - subtopic={subtopic}")
                # json-encode schema
                schema_enc = self.__encode_json(schema)
                # write encoded schema to output
                lines.append(b"schema_" + subtopic + b"||" + schema_enc)
                self.logger.debug(f"encode(): wrote schema to payload - subtopic={subtopic}")
            # avro-encode content
            content_enc = self.__encode_avro(subtopic, content)
            # write encoded content to output
            lines.append(prefix + subtopic + b"||" + content_enc)
        else:
            # json-encode content
            content_enc = self.__encode_json(content)
            # write encoded content to output
            lines.append(topic + b"||" + content_enc)
        return lines

    def decode(
        self,
        payload: bytes,
    ) -> Tuple[bytes, dict] | None:
        self.logger.debug(f"decode(): payload={payload}")
        topic, content_enc = payload.split(b"||", maxsplit=1)

        if len(content_enc) == 0:
            self.logger.info(f"decode(): got empty message - topic={topic}")
            return topic, {}
        elif topic.startswith(b"schema_"):
            subtopic = topic[7:]
            schema = self.__decode_json(content_enc)
            self.logger.debug(f"decode(): read schema from payload - subtopic={subtopic}")
            self.__cache_schema(subtopic, avro.schema.make_avsc_object(schema))
            self.logger.debug(f"decode(): assigned schema - subtopic={subtopic}")
            return None
        elif topic.startswith(prefix):
            subtopic = topic[len(prefix) :]
            if subtopic not in self.__schema_registry:
                self.logger.error(f"decode(): no schema - subtopic={subtopic}")
                return None
            schema = self.__schema_registry[subtopic]
            content = self.__decode_avro(content_enc, schema)
            # postprocessing
            content = self.__postprocess(content)
            return topic, content
        else:
            # decode using json backend
            content = self.__decode_json(content_enc)
            return topic, content



import pylsl
import xmltodict
from dfds.typing import Stream, dtype_map_inv


def stream_to_stream_info(
    stream: Stream,
) -> pylsl.StreamInfo:
    stream_info = pylsl.StreamInfo(
        name=stream.name,
        type=stream.unit,
        channel_count=len(stream.fields),
        nominal_srate=stream.frequency,
        channel_format=pylsl.cf_float32,
    )
    n_root = stream_info.desc()
    # description
    n_root.append_child_value("description", stream.description)
    # fields
    n_fields = n_root.append_child("fields")
    for k, v in stream.fields.items():
        n_fields_k = n_fields.append_child(k)
        n_fields_k.append_child_value("name", v.name)
        n_fields_k.append_child_value("description", v.description)
        n_fields_k.append_child_value("dtype", dtype_map_inv[v.dtype])
    # index
    n_index = n_root.append_child("index")
    for k, v in stream.index.items():
        n_index_k = n_index.append_child(k)
        n_index_k.append_child_value("name", v.name)
        n_index_k.append_child_value("description", v.description)
        n_index_k.append_child_value("dtype", dtype_map_inv[v.dtype])
    # attrs
    n_attrs = n_root.append_child("attrs")
    for k, v in stream.attrs.items():
        n_attrs.append_child_value(k, v)
    # device
    assert stream.node
    n_node = n_root.append_child("node")
    if stream.node.device:
        device = stream.node.device
        n_node_device = n_node.append_child("device")
        n_node_device.append_child_value("category", device.category)
        n_node_device.append_child_value("manufacturer", device.manufacturer)
        n_node_device.append_child_value("model", device.model)
    return stream_info


def stream_info_to_stream(
    stream_info: pylsl.StreamInfo,
) -> Stream:
    # read stream info in dict format
    inlet = pylsl.StreamInlet(stream_info)
    stream = stream_inlet_to_stream(inlet)
    inlet.close_stream()
    return stream


def stream_inlet_to_stream(
    stream_inlet: pylsl.StreamInlet,
) -> Stream:
    # read stream info in dict format
    info_xml = stream_inlet.info().as_xml()

    # parse stream info
    info: dict = xmltodict.parse(info_xml)["info"]
    desc: dict = info["desc"]
    desc["name"] = info["name"] or ""
    desc["unit"] = info["type"] or ""
    desc["frequency"] = info["nominal_srate"] or ""
    desc["@node"] = desc.pop("node")
    # generate stream object from dict
    return Stream(**desc)



import logging
from abc import ABC, abstractmethod
from typing import Tuple


class Codec(ABC):
    """
    Base Class for Serialization / Deserialization

    """

    def __init__(
        self,
    ):
        self.logger = logging.getLogger(__name__)
        self.active = False

    @abstractmethod
    def encode(
        self,
        topic: bytes,
        content: dict,
    ) -> bytes | list[bytes]:
        raise NotImplementedError()

    @abstractmethod
    def decode(
        self,
        payload: bytes,
    ) -> Tuple[bytes, dict] | None:
        raise NotImplementedError()



import asyncio
import multiprocessing
from threading import Thread

from datamux.api import DataMuxAPI

from .rpc import create_rpc_server
from .topics import *


class DataMuxServer:
    """
    Serves the DataMux API for Remote Use

    """

    def __init__(
        self,
        rpc_name: str,
        codec_name: str,
    ) -> None:
        self.active = False
        self.api_in = asyncio.Queue()
        self.api_out = asyncio.Queue()
        self.data_q = multiprocessing.Queue()
        # rpc module
        self.rpc = create_rpc_server(
            name=rpc_name,
            codec_name=codec_name,
            incoming=self.api_in,
            outgoing=self.api_out,
        )
        # api module
        self.api = DataMuxAPI()

    async def handle_requests(
        self,
    ):
        """
        Handle Requests sent to DataMuxServer

        Args:
            topic (bytes): topic of the message
            content (dict): content of the message
        """

        while self.active:
            topic, content, id = await self.api_in.get()
            transform = lambda retval: [retval, id]

            # LIVE MODE (LSL -> Queue) =================================================================================================
            if topic == TOPIC_LIST_LIVE_STREAMS:
                streams = self.api.list_live_streams()
                retval = [s.model_dump() for s in streams]
            elif topic == TOPIC_READ_LIVE_STREAM:
                stream_name = content["stream_name"]
                attrs = content["attrs"]
                ack = self.api.read_live_stream(stream_name, attrs, self.data_q, transform)
                retval = ack.model_dump()
            # REPLAY MODE (File -> Queue) ==============================================================================================
            elif topic == TOPIC_LIST_COLLECTIONS:
                collections = self.api.list_collections()
                retval = [c.model_dump() for c in collections]
            elif topic == TOPIC_LIST_COLLECTION_STREAMS:
                collec_name = content["collection_name"]
                streams = self.api.list_collection_streams(collec_name)
                retval = [s.model_dump() for s in streams]
            elif topic == TOPIC_REPLAY_COLLECTION_STREAM:
                collec_name = content["collection_name"]
                stream_name = content["stream_name"]
                attrs = content["attrs"]
                ack = self.api.replay_collection_stream(collec_name, stream_name, attrs, self.data_q, transform)
                retval = ack.model_dump()
            # RESTREAM MODE (File -> LSL) ==============================================================================================
            elif topic == TOPIC_PUBLISH_COLLECTION_STREAM:
                collection_name = content["collection_name"]
                stream_name = content["stream_name"]
                attrs = content["attrs"]
                ack = self.api.publish_collection_stream(collection_name, stream_name, attrs)
                retval = ack.model_dump()
            # ACTIONS ==================================================================================================================
            elif topic == TOPIC_STOP_TASK:
                randseq = content["randseq"]
                ack = self.api.stop_task(randseq)
                retval = ack.model_dump()
            # FALLBACK =================================================================================================================
            else:
                retval = dict(error="Unknown Request")

            msg = [topic] + transform(retval)
            self.api_out.put_nowait(msg)

    def requeue(
        self,
        loop: asyncio.AbstractEventLoop,
    ):
        while self.active:
            msg = self.data_q.get()
            loop.call_soon_threadsafe(self.api_out.put_nowait, msg)

    async def start(
        self,
        host: str,
        port: int,
    ):
        self.active = True
        loop = asyncio.get_running_loop()
        Thread(None, self.requeue, None, (loop,), daemon=True).start()
        asyncio.create_task(self.handle_requests())
        asyncio.create_task(self.rpc.start(host, port))
        flag = asyncio.Future()
        await flag

    async def stop(
        self,
    ):
        self.active = False
        await self.rpc.stop()



#!/usr/bin/env python3

import asyncio
import logging
import timeit

from rich.logging import RichHandler

import datamux.util as util
from datamux.remote.api import DataMuxRemoteAPI


async def main():
    N = 10
    P = 8000
    t_replay = []
    t_relay = []
    server_host = "localhost"
    server_port = 3300

    api = DataMuxRemoteAPI(rpc_name="websocket", codec_name="avro")
    await api.connect(server_host, server_port)

    for _ in range(N):
        # test 1 - collections
        logger.info("listing collections")
        collections = await api.list_collections()
        logger.info(f"received collections: {collections}")

        # test 2 - collection streams
        logger.info("listing collection streams - adhd_sin")
        collection_streams = await api.list_collection_streams("adhd_sin")
        logger.info(f"received collection_streams: {collection_streams}")

        # test 3 - replay collection stream
        logger.info("replaying collection stream")
        collection_name = "adhd_sin"
        stream_name = "Gaze"
        attrs = dict(
            {
                "noise": "0",
                "participant": "12",
                "question": "10",
            }
        )
        sink = asyncio.Queue()
        ack = await api.replay_collection_stream(collection_name, stream_name, attrs, sink)
        assert ack.randseq is not None
        logger.info(f"received ack for collection stream: {ack}")
        # print P points or until EOF
        startTime = timeit.default_timer()
        for i in range(P):
            item = await sink.get()
            logger.info(item)
            if item == util.END_OF_STREAM:
                break
            if i == 0:
                startTime = timeit.default_timer()
        endTime = timeit.default_timer()
        t_replay.append(endTime - startTime)
        await api.stop_task(ack.randseq)

        # test 4 - make LSL stream from a collection stream
        logger.info("creating LSL stream")
        status = await api.publish_collection_stream(collection_name, stream_name, attrs)
        logger.info(f"created LSL stream: {status}")
        assert status
        await asyncio.sleep(5)

        # test 5 - list all LSL streams
        logger.info("listing LSL streams")
        live_streams = await api.list_live_streams()
        logger.info(f"got LSL streams: {live_streams}")

        # test 6 - relay LSL stream
        logger.info("relaying LSL stream")
        assert len(live_streams) > 0
        ls = live_streams[0]
        sink = asyncio.Queue()
        ack = await api.read_live_stream(ls.name, ls.attrs, sink)
        assert ack.randseq is not None
        logger.info(f"received ack for live stream: {ack}")
        # print P points or until EOF
        startTime = timeit.default_timer()
        for i in range(P):
            item = await sink.get()
            logger.info(item)
            if item == util.END_OF_STREAM:
                break
            if i == 0:
                startTime = timeit.default_timer()
        endTime = timeit.default_timer()
        t_relay.append(endTime - startTime)
        await api.stop_task(ack.randseq)

    print("Elapsed Time (Replay): ", t_replay)
    print("Elapsed Time (Relay): ", t_relay)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(message)s", datefmt="[%X]", handlers=[RichHandler()])
    logger = logging.getLogger(__name__)
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        logger.warning("Interrupt received, shutting down.")



import asyncio
import random
from functools import partial, wraps
from typing import Any, Dict, Generator

from pydantic import BaseModel

DICT = Dict[str, Any]
DICT_GENERATOR = Generator[DICT, None, None]
END_OF_STREAM = {}  # NOTE do not change


def asyncify(func, executor):
    @wraps(func)
    async def run(*args, **kwargs):
        return asyncio.get_event_loop().run_in_executor(executor, partial(func, *args, **kwargs))

    return run


def gen_randseq(length: int = 5) -> str:
    options = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    return "".join(random.choice(options) for x in range(length))


class StreamAck(BaseModel):
    status: bool
    randseq: str | None = None



import json
import logging
from typing import Tuple

from .codec import Codec


class JSONCodec(Codec):
    """
    JSON Codec

    """

    logger = logging.getLogger(__name__)

    def encode(
        self,
        topic: bytes,
        content: dict,
    ) -> bytes:
        self.logger.debug(f"encode(): topic={topic}, content={content}")
        if len(content) == 0:
            content_enc = b""
        else:
            content_str = json.dumps(content)
            content_enc = content_str.encode()
        payload = topic + b"||" + content_enc
        return payload

    def decode(
        self,
        payload: bytes,
    ) -> Tuple[bytes, dict]:
        self.logger.debug(f"decode(): payload={payload}")
        topic, content_enc = payload.split(b"||", maxsplit=1)
        content = {}
        if len(content_enc) > 0:
            content_str = content_enc.decode()
            content = json.loads(content_str)
        return topic, content



#!/usr/bin/env python3

"""
This command-line executable simulates the datasets
stream of data in real-time, and generates a
datasets-stream based on the device specifications.

It can be used to replay an experimental setup from
the datasets that's already collected.
"""

import argparse
import asyncio
import logging
import threading
from multiprocessing import Queue

from dfds.typing import Stream
from rich.logging import RichHandler

import datamux.util as util
from datamux.api import DataMuxAPI

DIGIT_CHARS = "0123456789"
SHUTDOWN_FLAG = threading.Event()
logger = logging.getLogger(__name__)

api = DataMuxAPI()


def log_sink(sink: Queue):
    while True:
        value = sink.get()
        logging.info(value)
        if value == util.END_OF_STREAM:
            break


async def begin_streaming(collection_name: str, streams: list[Stream]):
    logger.info(f"Started data streaming")

    threads: list[threading.Thread] = []

    for stream in streams:
        node = stream.node
        assert node is not None
        device = node.device
        assert device is not None
        source_id = util.gen_randseq()
        logger.info(f"Source [{source_id}]: {device.model}, {device.manufacturer} ({device.category})")
        sink = Queue()
        api.replay_collection_stream(collection_name, stream.name, stream.attrs, sink)
        threads.append(threading.Thread(target=log_sink, args=(sink,)))
        logger.info(f"Source [{source_id}]: started")

    for thread in threads:
        thread.join()

    logger.info(f"Ended data streaming")


def main():
    # get default args
    # create parser and parse args
    parser = argparse.ArgumentParser(prog="replay.py")
    # required args
    parser.add_argument("--collection-name", "-n", required=True)
    parser.add_argument("--data-dir", "-d", required=True)
    parser.add_argument("--meta-dir", "-m", required=True)
    # optional args
    parser.add_argument("--attributes", "-a", required=False)
    # parse all args
    known_args, args = parser.parse_known_args()
    collection_name = known_args.collection_name
    # print args
    logger.info(f"Collection Name: {collection_name}")
    logger.info(f"Data Directory: {known_args.data_dir}")
    logger.info(f"Meta Directory: {known_args.meta_dir}")
    # get dataset spec
    streams = api.list_collection_streams(collection_name)
    # get data sources
    assert len(streams) > 0, f"Dataset does not have data sources"
    # spawn a worker thread for streaming
    logger.info("=== Begin streaming ===")
    try:
        asyncio.run(begin_streaming(collection_name, streams))
    except (KeyboardInterrupt, InterruptedError):
        logger.info("Interrupt received. Ending all stream tasks..")
    logger.info("All streaming tasks ended")


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(message)s", datefmt="[%X]", handlers=[RichHandler()])
    try:
        main()
    except AssertionError as e:
        logger.error(f"Error: {e}")



#!/usr/bin/env python
import asyncio
import io
import logging
from typing import Any, Dict, Optional

import dfds
import msgpack
import numpy as np
import pylsl
import zmq
import zmq.asyncio
from PIL import Image
from rich.logging import RichHandler

from . import Interface

ZMQ_REQ = zmq.REQ
ZMQ_SUB = zmq.SUB
ZMQ_SUBSCRIBE = zmq.SUBSCRIBE
CONF_THRESHOLD = 0.8


class Connector(Interface):
    """
    Pupil Core -> LSL Connector (via ZMQ)

    """

    def __init__(
        self,
        host: str,
        port: int,
    ) -> None:
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self.host = host
        self.port = port
        self.parser = dfds.Parser()
        self.node = self.parser.get_node_metadata("repository/pupil_core.node.json")
        self.ctx = zmq.asyncio.Context()
        self.req = self.ctx.socket(ZMQ_REQ)
        self.sub = self.ctx.socket(ZMQ_SUB)
        self.topics = [
            "frame.world",  # camera stream (eye, world)
            "gaze.3d.0.",  # where you're looking at
            "gaze.3d.1.",  # where you're looking at
            # 'pupil', # information about pupil (such as diameter, dilation)
            "fixations",  # starts sending data when it detects fixations (continuously sends until next fixation)
            # 'notify',  # notifications
            # 'logging'  # for logging
        ]
        self.stream_idx = {
            "frame.world": 0,
            "gaze.3d.0.": 1,
            "gaze.3d.1.": 2,
            "fixations": 3,
        }
        self.outlets: Dict[str, pylsl.StreamOutlet] = {}
        self.sub_port: int

    async def get_sub_port(
        self,
    ) -> int:
        self.req.connect(f"tcp://{self.host}:{self.port}")
        self.logger.debug("Opened REQ socket.")
        self.req.send_string("SUB_PORT")
        sub_port = int(await self.req.recv_string())
        self.logger.debug(f"Received SUB_PORT. {self.sub_port}")
        return sub_port

    def get_outlet(
        self,
        stream_id: str,
    ) -> pylsl.StreamOutlet:
        self.outlets[stream_id] = self.outlets.get(
            stream_id,
            dfds.create_outlet(
                stream_id=stream_id,
                stream=self.node.outputs[stream_id],
            ),
        )
        return self.outlets[stream_id]

    def subscribe(
        self,
    ):
        self.sub.connect(f"tcp://{self.host}:{self.sub_port}")
        self.logger.debug("Opened SUB socket")
        for sub_id in self.topics:
            self.sub.setsockopt_string(ZMQ_SUBSCRIBE, sub_id)
            self.logger.debug(f"Subscribed to: {sub_id}")
        self.logger.debug("Subscription Completed.")

    def emit(
        self,
        topic: str,
        message: dict,
        image: Optional[np.ndarray],
    ) -> Any:
        ts = message["timestamp"]
        if topic.startswith("gaze.3d."):
            # decode message
            [gaze_pos_x, gaze_pos_y] = message["norm_pos"]
            gaze_conf = message["confidence"]
            # push to LSL stream if data point has acceptable confidence
            if gaze_conf > CONF_THRESHOLD:
                self.get_outlet(topic[:10]).push_sample([gaze_pos_x, gaze_pos_y, gaze_conf], ts)
        elif topic.startswith("fixations"):
            # decode message
            [fxn_pos_x, fxn_pos_y] = message["norm_pos"]
            fxn_dispersion = message["dispersion"]
            fxn_d = message["duration"]
            fxn_conf = message["confidence"]
            # push to LSL stream if data point has acceptable confidence
            if fxn_conf > CONF_THRESHOLD:
                self.get_outlet(topic[:9]).push_sample([fxn_pos_x, fxn_pos_y, fxn_dispersion, fxn_d, fxn_conf], ts)
        elif image is not None:
            # push image data to LSL stream
            [r, g, b] = image
            # TODO only sending (0,0) pixel for now, fix to send all pixels
            self.get_outlet(topic[:11]).push_sample([r[0][0], g[0][0], b[0][0]], ts)

    async def run(
        self,
    ):
        self.sub_port = await self.get_sub_port()
        self.subscribe()
        self.logger.debug("Started Data Stream. Use SIGINT to terminate.")
        while True:
            try:
                # receive raw data (bytes)
                [topic, payload, *extra] = await self.sub.recv_multipart()
                topic = topic.decode()
                image = None
                # world camera
                if topic.startswith("frame.world"):
                    image = np.transpose(np.asarray(Image.open(io.BytesIO(extra[0])).convert("RGB")))
                # payload
                payload = msgpack.loads(payload)
                assert type(payload) == dict
                self.emit(topic, payload, image)
            except KeyboardInterrupt:
                self.logger.debug("Received Interrupt. Stopping...")
                break
        self.logger.debug("Stopped.")


async def main():
    logging.basicConfig(level=logging.DEBUG, format="%(message)s", datefmt="[%X]", handlers=[RichHandler()])
    connector = Connector(host="127.0.0.1", port=50020)
    await connector.run()


if __name__ == "__main__":
    asyncio.run(main())



import logging
import random
import time
from multiprocessing import Queue
from threading import Event, Thread
import timeit
import numpy as np

import pylsl
from dfds import Parser
from dfds.typing import Collection, Config, Stream

import datamux.util as util

from . import Reader
from .util import stream_to_stream_info


# TODO change restreaming from pointwise to groupwise
class CollectionReader(Reader):
    """
    Stream Reader for DFDS Collections.

    Functions provided:

    * list_collections() - return a list of currently available collections

    * refresh_collections() - refresh the list of currently available collections

    * list_streams(collection_name) - list the streams in a collection

    * replay(collection_name, stream_name, queue) - replay data from a stream in a collection

    * restream(collection_name, stream_name) - restream (via lsl) data from a stream in a collection

    """

    __parser = Parser()
    __collections: list[Collection] = []
    logger = logging.getLogger(__name__)

    def __init__(
        self,
        config: Config,
    ) -> None:
        super().__init__()
        self.config = config

    def list_collections(
        self,
    ) -> list[Collection]:
        return self.__collections

    def refresh_collections(
        self,
    ) -> None:
        self.__collections.clear()
        for fp in self.config.meta_dir.glob("*.collection.json"):
            collection = self.__parser.get_collection_metadata(fp.as_posix())
            self.__collections.append(collection)

    def list_streams(
        self,
        collection_name: str,
    ) -> list[Stream]:
        collection = [c for c in self.__collections if c.name == collection_name][0]
        streams = []
        dataloader = collection.dataloader(self.config)
        for attrs in dataloader.ls():
            for stream_id, stream in collection.streams.items():
                stream = stream.model_copy(deep=True)
                stream.attrs.update(attrs)
                stream.attrs.update({"id": stream_id})
                streams.append(stream)
        return streams

    def replay(
        self,
        collection_name: str,
        stream_name: str,
        attrs: dict,
        queue: Queue,
        transform,
        flag,
    ) -> None:
        collection = [c for c in self.__collections if c.name == collection_name][0]
        stream = [s for s in collection.streams.values() if s.name == stream_name][0]
        stream.attrs.update(attrs, dfds_mode="replay")
        thread = Thread(None, self.replay_coro, stream_name, (collection, stream, queue, transform, flag), daemon=True)
        thread.start()

    def restream(
        self,
        collection_name: str,
        stream_name: str,
        attrs: dict,
    ) -> None:
        collection = [c for c in self.__collections if c.name == collection_name][0]
        stream = [s for s in collection.streams.values() if s.name == stream_name][0]
        stream.attrs.update(attrs, dfds_mode="restream")
        thread = Thread(None, self.restream_coro, stream_name, (collection, stream), daemon=True)
        thread.start()

    def replay_coro(
        self,
        collection: Collection,
        stream: Stream,
        queue: Queue,
        transform,
        flag: Event,
        strict_time: bool = True,
    ):
        freq = stream.frequency
        if freq <= 0:
            freq = random.randint(1, 50)  # assign a random frequency between 0 and 50
        dt = 1 / freq
        index_cols = list(stream.index)
        value_cols = list(stream.fields)

        attrs, data = collection.dataloader(self.config).read(stream.attrs)
        stream.attrs.update(attrs)

        # termination indicator
        eof = util.END_OF_STREAM
        if transform is not None:
            eof = transform(eof)

        # preprocessing
        data = data.sort_values(index_cols, ascending=True)
        # if index not present, create one with given frequency
        for col in index_cols:
            if col not in data.columns:
                data[col] = np.arange(len(data)) / freq
        # TODO time to seconds
        primary_index = index_cols[0]
        dt_true = float(data.iloc[1][primary_index] - data.iloc[0][primary_index])
        si_scales = np.array([1, 1e3, 1e6, 1e9], dtype=int)
        scale = si_scales[np.abs(si_scales - dt_true / dt).argmin()]
        data[primary_index] = data[primary_index] / scale

        # replay each record
        self.logger.info(f"replay started")
        t0, T0 = None, None
        for _, record in data.iterrows():
            # terminating condition
            if flag.is_set():
                self.logger.info(f"replay stop requested")
                break

            # create record
            index = {k: record[k] for k in index_cols}
            value = {k: record[k] for k in value_cols}
            msg = dict(index=index, value=value)
            if transform is not None:
                msg = transform(msg)

            # wait until time requirements are met
            if strict_time:
                if t0 is None or T0 is None:
                    t0, T0 = timeit.default_timer(), index[index_cols[0]]
                else:
                    ti, Ti = timeit.default_timer(), index[index_cols[0]]
                    dt = (Ti - T0) - (ti - t0)
                    if dt > 0:
                        time.sleep(dt)
            else:
                time.sleep(dt)

            # send record
            queue.put_nowait(msg)

        queue.put_nowait(eof)
        self.logger.info(f"replay ended")

    def restream_coro(
        self,
        collection: Collection,
        stream: Stream,
    ):
        freq = stream.frequency
        if freq <= 0:
            freq = random.randint(1, 50)  # assign a random frequency between 0 and 50
        dt = 1 / freq
        index_cols = list(stream.index)
        value_cols = list(stream.fields)

        attrs, data = collection.dataloader(self.config).read(stream.attrs)
        stream.attrs.update(attrs)

        outlet = pylsl.StreamOutlet(stream_to_stream_info(stream))
        num_samples = data.shape[0]
        current_index = 0

        # restream each record
        started = False
        self.logger.info(f"started restream: dt={dt:.4f}, n={num_samples}")
        while current_index < num_samples:
            index = data[current_index][index_cols]
            value = data[current_index][value_cols]
            if outlet.have_consumers():
                started = True
                x = value.tolist()
                t = index[0]  # FIXME currently only supports 1D index
                outlet.push_sample(x, t)
                current_index += 1
            elif started:
                # if no consumers after starting, stop restreaming
                break
            time.sleep(dt)
        self.logger.info(f"ended restream")



import asyncio
import logging
from collections import defaultdict

from dfds.typing import Collection, Stream

from datamux.api import StreamAck

from .rpc import create_rpc_client
from .topics import *


class DataMuxRemoteAPI:
    """
    Remote API for DataMux.

    It provides three modes of execution.

    * **Listen** mode: read live data streams (LSL) on the local network.
    * **Replay** mode: replay datasets from storage to mimic a live data stream.
    * **Guided** mode: simulate random/guided data as a data stream.

    """

    def __init__(
        self,
        rpc_name: str,
        codec_name: str,
    ) -> None:
        """
        Create Remote API instance.

        Args:
            rpc_name (str): method of RPC. supports {websocket, direct}.
            codec_name (str): name of codec. supports {json, avro}.
        """
        self.active = False
        # queues - sending
        self.outgoing = asyncio.Queue()
        # queues - receiving
        self.incoming = asyncio.Queue()
        # rpc module
        self.rpc = create_rpc_client(
            name=rpc_name,
            codec_name=codec_name,
            incoming=self.incoming,
            outgoing=self.outgoing,
        )
        # request handling module
        self.handlers: dict[bytes, asyncio.Queue] = defaultdict(lambda: asyncio.Queue())
        self.logger = logging.getLogger(__name__)

    async def __handle_incoming__(
        self,
    ):
        while self.active:
            topic, content = await self.incoming.get()
            self.logger.debug(f"<: {topic}: {content}")
            await self.handlers[topic].put(content)

    async def __send_await__(
        self,
        topic: bytes,
        content: dict,
    ):
        self.logger.debug(f">: {topic}: {content}")
        await self.outgoing.put((topic, content))
        return self.handlers[topic]

    async def connect(
        self,
        server_host: str,
        server_port: int,
    ):
        """
        Establish remote connection.

        Args:
            server_host (str): Hostname of running server
            server_port (int): Port of running server
        """
        self.active = True
        await self.rpc.connect(server_host, server_port)
        asyncio.create_task(self.__handle_incoming__())

    async def disconnect(
        self,
    ):
        """
        Close remote connection.

        """
        self.active = False

    async def list_collections(
        self,
    ) -> list[Collection]:
        """
        List all collections.

        Returns:
            list[Collection]: list of available collections.
        """
        topic = TOPIC_LIST_COLLECTIONS
        content: dict[str, str] = {}
        result = await self.__send_await__(topic, content)
        items = await result.get()
        collections = [Collection(**item) for item in items]
        return collections

    async def list_collection_streams(
        self,
        collection_name: str,
    ) -> list[Stream]:
        """
        List all streams in a collection.

        Args:
            collection_name (str): name of collection.

        Returns:
            list[Stream]: list of streams in collection.
        """
        topic = TOPIC_LIST_COLLECTION_STREAMS
        content = dict(collection_name=collection_name)
        result = await self.__send_await__(topic, content)
        items = await result.get()
        streams = [Stream(**item) for item in items]
        return streams

    async def replay_collection_stream(
        self,
        collection_name: str,
        stream_name: str,
        attrs: dict[str, str],
        sink: asyncio.Queue,
    ) -> StreamAck:
        """
        Replay a collection-stream into a given queue.

        Args:
            collection_name (str): name of collection.
            stream_name (str): name of stream in collection.
            attrs (dict): attributes specifying which recording to replay.
            sink (asyncio.Queue): destination to buffer replayed data.

        Returns:
            StreamAck: status and reference information.
        """
        topic = TOPIC_REPLAY_COLLECTION_STREAM
        content = dict(
            collection_name=collection_name,
            stream_name=stream_name,
            attrs=attrs,
        )
        result = await self.__send_await__(topic, content)
        info = await result.get()
        ack = StreamAck(**info)
        assert ack.randseq is not None
        stream_topic = ack.randseq.encode()
        self.handlers[stream_topic] = sink
        return ack

    async def publish_collection_stream(
        self,
        collection_name: str,
        stream_name: str,
        attrs: dict[str, str],
    ) -> StreamAck:
        """
        Publish a collection-stream as a LSL stream.

        Args:
            collection_name (str): name of collection.
            stream_name (str): name of stream in collection.
            attrs (dict): attributes specifying which recording to publish.

        Returns:
            StreamAck: status and reference information.
        """
        topic = TOPIC_PUBLISH_COLLECTION_STREAM
        content = dict(
            collection_name=collection_name,
            stream_name=stream_name,
            attrs=attrs,
        )
        result = await self.__send_await__(topic, content)
        info = await result.get()
        ack = StreamAck(**info)
        return ack
    
    async def stop_task(
        self,
        randseq: str,
    ) -> StreamAck:
        topic = TOPIC_STOP_TASK
        content = dict(randseq=randseq)
        result = await self.__send_await__(topic, content)
        info = await result.get()
        ack = StreamAck(**info)
        return ack

    async def list_live_streams(
        self,
    ) -> list[Stream]:
        """
        List all live streams.

        Returns:
            list[Stream]: list of available live streams.
        """
        topic = TOPIC_LIST_LIVE_STREAMS
        content: dict[str, str] = {}
        result = await self.__send_await__(topic, content)
        items = await result.get()
        streams = [Stream(**item) for item in items]
        return streams

    async def read_live_stream(
        self,
        stream_name: str,
        attrs: dict,
        sink: asyncio.Queue,
    ) -> StreamAck:
        """
        Read data from a live stream (LSL) into a given queue.

        Args:
            stream_name (str): name of live stream.
            attrs (dict): attributes specifying which live stream to read.
            sink (asyncio.Queue): destination to buffer replayed data.

        Returns:
            StreamAck: status and reference information.
        """
        topic = TOPIC_READ_LIVE_STREAM
        content = dict(stream_name=stream_name, attrs=attrs)
        result = await self.__send_await__(topic, content)
        info = await result.get()
        ack = StreamAck(**info)
        assert ack.randseq is not None
        stream_topic = ack.randseq.encode()
        self.handlers[stream_topic] = sink
        return ack



#!/usr/bin/env python3
import codecs
import logging
import re
import socket

import dfds
import pylsl
from rich.logging import RichHandler

from . import Interface


class E4ServerState:
    """
    Server States
    """

    NEW = "new"
    WAITING = "waiting"
    NO_DEVICES = "no_devices"
    DEVICES_FOUND = "devices_found"
    CONNECTED_TO_DEVICE = "connected"
    READY_TO_SUBSCRIBE = "ready_to_subscribe"
    SUBSCRIBE_COMPLETED = "subscribe completed"
    STREAMING = "streaming"


class E4SSCommand:
    """
    Server Commands
    """

    DEVICE_DISCOVER_LIST = "device_discover_list"
    DEVICE_CONNECT_BTLE = "device_connect_btle"
    DEVICE_DISCONNECT_BTLE = "device_disconnect_btle"
    DEVICE_LIST = "device_list"
    DEVICE_CONNECT = "device_connect"
    DEVICE_DISCONNECT = "device_disconnect"
    DEVICE_SUBSCRIBE = "device_subscribe"
    PAUSE = "pause"


class Connector(Interface):
    """
    Empatica E4 -> LSL Connector (via E4 Streaming Server)
    """

    buffer_size: int = 4096
    logger = logging.getLogger(__name__)

    def __init__(
        self,
    ) -> None:
        super().__init__()
        self.parser: dfds.Parser = dfds.Parser()
        self.node = self.parser.get_node_metadata("repository/empatica_e4.node.json")
        self.num_subs: int = 0
        self.device_id: str
        self.outlets: dict = {}
        self.server_state = E4ServerState.NEW
        self.patterns = [
            # (typ, cmd, arg, data)
            re.compile(
                r"^(?:(R)\s)?(device_discover_list|device_connect_btle|device_disconnect_btle|device_list|device_connect|device_disconnect|pause)(?:\s([\.\w]+))?(?:\s?(.+?))?$"
            ),
            # (typ, cmd, sid, arg, data)
            re.compile(r"^(?:(R)\s)?(device_subscribe)(?:\s(\w+))?(?:\s([\.\w]+))?(?:\s?(.+?))?$"),
            # (sid, time, data)
            re.compile(
                r"^(\w+)\s([\d.]+)\s?(.*)$",
            ),
        ]

    def encode(self, s_: str) -> bytes:
        return codecs.encode(s_ + "\r\n")

    def get_outlet(self, stream_id: str) -> pylsl.StreamOutlet:
        self.outlets[stream_id] = self.outlets.get(
            stream_id,
            dfds.create_outlet(
                stream_id=stream_id,
                stream=self.node.outputs[stream_id],
            ),
        )
        return self.outlets[stream_id]

    def start(self):
        # Create socket connection
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((socket.gethostname(), 28000))
        while True:
            self.handle_outgoing_msgs(sock)
            self.handle_incoming_msgs(sock)

    def handle_outgoing_msgs(
        self,
        s: socket.socket,
    ):
        if self.server_state == E4ServerState.NEW:
            # request devices list
            self.logger.debug("Getting list of devices...")
            s.send(self.encode(E4SSCommand.DEVICE_LIST))
            self.server_state = E4ServerState.WAITING
        elif self.server_state == E4ServerState.NO_DEVICES:
            self.logger.debug("No devices found!")
            exit(1)
        elif self.server_state == E4ServerState.DEVICES_FOUND:
            # connect to device
            self.logger.debug("Connecting to device...")
            s.send(self.encode("%s %s" % (E4SSCommand.DEVICE_CONNECT, self.device_id)))
            self.server_state = E4ServerState.WAITING
        elif self.server_state == E4ServerState.CONNECTED_TO_DEVICE:
            # pause streaming initially
            self.logger.debug("Initializing...")
            s.send(self.encode("%s ON" % E4SSCommand.PAUSE))
            self.server_state = E4ServerState.WAITING
        elif self.server_state == E4ServerState.READY_TO_SUBSCRIBE:
            # subscribe to streams
            stream = sorted(self.node.outputs)[self.num_subs]
            self.logger.debug("Subscribing to stream: %s" % stream)
            s.send(self.encode("%s %s ON" % (E4SSCommand.DEVICE_SUBSCRIBE, stream)))
            self.server_state = E4ServerState.WAITING
        elif self.server_state == E4ServerState.SUBSCRIBE_COMPLETED:
            # begin streaming data
            self.logger.debug("Requesting data")
            s.send(self.encode("%s OFF" % E4SSCommand.PAUSE))
            self.server_state = E4ServerState.STREAMING

    def handle_incoming_msgs(
        self,
        s: socket.socket,
    ):
        # pull messages from socket
        buffer: str = codecs.decode(s.recv(self.buffer_size))
        for message in [*map(str.strip, buffer.split("\r\n"))]:
            # parse message
            typ: str = ""
            cmd: str = ""
            sid: str = ""
            arg: str = ""
            data: str = ""
            if (match := self.patterns[0].match(message)) is not None:
                (typ, cmd, arg, data) = match.groups()
            elif (match := self.patterns[1].match(message)) is not None:
                (typ, cmd, sid, arg, data) = match.groups()
            elif (match := self.patterns[2].match(message)) is not None:
                (sid, arg, data) = match.groups()
                sid = sid[3:].lower()  # drop E4_ prefix and lowercase
            else:
                raise ValueError(f"Unknown message: {message}")

            # Data Stream
            if typ == "" and self.server_state == E4ServerState.STREAMING:
                outlet = self.get_outlet(sid)
                # only parse command if outlet has consumers
                if outlet.have_consumers():
                    outlet.push_sample([*map(float, data.split())] or [0], float(arg))

            # DEVICE_LIST response
            elif cmd == E4SSCommand.DEVICE_LIST:
                device_ids = [x.split()[0] for x in data.strip("| ").split("|")]
                num = len(device_ids)
                assert num == int(arg)
                self.logger.debug(f"device(s) found: {num}")

                if num == 0:
                    self.server_state = E4ServerState.NO_DEVICES
                    raise ValueError(f"No devices found: {num}")

                self.server_state = E4ServerState.DEVICES_FOUND

                # choose the only device
                if num == 1:
                    return device_ids[0]

                # prompt user to select device
                while True:
                    device_id = input(f"Enter device id: ")
                    if device_id in device_ids:
                        return device_id
                    self.logger.debug(f"Invalid device id: {device_id}")

            # DEVICE_CONNECT response
            elif cmd == E4SSCommand.DEVICE_CONNECT:
                if arg == "ERR":
                    raise ValueError(f"Error connecting to device: {data}")
                elif arg == "OK":
                    self.logger.debug("Connected to device")
                    self.server_state = E4ServerState.CONNECTED_TO_DEVICE

            # PAUSE response
            elif cmd == E4SSCommand.PAUSE:
                if arg == "ERR":
                    raise ValueError(f"Error pausing streaming: {data}")
                elif arg == "ON":
                    self.logger.debug("Streaming on hold")
                    self.server_state = E4ServerState.READY_TO_SUBSCRIBE
                elif arg == "OFF":
                    self.logger.debug("Streaming started")
                    self.server_state = E4ServerState.STREAMING

            # DEVICE SUBSCRIBE response
            elif cmd == E4SSCommand.DEVICE_SUBSCRIBE:
                if arg == "ERR":
                    raise ValueError(f"Error subscribing to: {sid} - {data}")
                elif arg == "OK":
                    self.logger.debug(f"Subscribed to: {sid}")
                    self.num_subs += 1
                    if self.num_subs == len(self.node.outputs):
                        self.server_state = E4ServerState.SUBSCRIBE_COMPLETED
                    else:
                        self.server_state = E4ServerState.READY_TO_SUBSCRIBE

            # default
            else:
                raise ValueError(f"Unknown message: {message}")


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG, format="%(message)s", datefmt="[%X]", handlers=[RichHandler()])
    connector = Connector()
    connector.start()



import asyncio
import logging

from websockets.client import WebSocketClientProtocol, connect

from . import RpcClient
from ..codec import create_codec


class WebsocketRPC(RpcClient):
    """
    WebSocket RPC Client

    """

    def __init__(
        self,
        codec_name: str,
        incoming: asyncio.Queue,
        outgoing: asyncio.Queue,
    ) -> None:
        self.active = False
        self.codec = create_codec(codec_name)
        self.incoming = incoming
        self.outgoing = outgoing
        self.logger = logging.getLogger(__name__)

    async def connect(
        self,
        server_host: str,
        server_port: int,
    ) -> None:
        self.active = True
        uri = f"ws://{server_host}:{server_port}/ws"
        self.websocket = await connect(uri)
        self.logger.info(f"Connected to WebSocket Server: {uri}")
        assert self.websocket is not None
        asyncio.create_task(self.__handle_outgoing__(self.websocket))
        asyncio.create_task(self.__handle_incoming__(self.websocket))

    async def disconnect(
        self,
    ):
        self.active = False
        await asyncio.create_task(self.websocket.close())

    async def __handle_outgoing__(
        self,
        websocket: WebSocketClientProtocol,
    ):
        while self.active and websocket.open:
            topic, content = await self.outgoing.get()
            self.logger.debug(f">: {topic}: {content}")
            msg: bytes | list[bytes] = self.codec.encode(topic, content)
            if isinstance(msg, bytes):
                await self.websocket.send(msg)
            if isinstance(msg, list):
                for frame in msg:
                    await self.websocket.send(frame)

    async def __handle_incoming__(
        self,
        websocket: WebSocketClientProtocol,
    ):
        while self.active and websocket.open:
            msg = await self.websocket.recv()
            assert isinstance(msg, bytes)
            if (val := self.codec.decode(msg)) is not None:
                topic, content = val
                self.logger.debug(f"<: {topic}: {content}")
                await self.incoming.put((topic, content))



from multiprocessing import Queue
from threading import Event

import dfds
from dfds.typing import Collection, Stream

from .readers import CollectionReader, NodeReader
from .util import StreamAck, gen_randseq

prefix = "d_"


class DataMuxAPI:
    """
    API for DataMux.

    It provides three modes of execution.

    * **Listen** mode: read live data streams (LSL) on the local network.
    * **Replay** mode: replay datasets from storage to mimic a live data stream.
    * **Guided** mode: simulate random/guided data as a data stream.

    """

    def __init__(self):
        """
        Create API instance.

        """
        self.config = dfds.load_config()
        self.reader_n = NodeReader()
        self.reader_c = CollectionReader(self.config)
        self.context: dict[str, Event] = {}

    def list_collections(
        self,
    ) -> list[Collection]:
        """
        List all collections.

        Returns:
            list[Collection]: list of available collections.
        """
        self.reader_c.refresh_collections()
        collections = self.reader_c.list_collections()
        return collections

    def list_collection_streams(
        self,
        collection_name: str,
    ) -> list[Stream]:
        """
        List all streams in a collection.

        Args:
            collection_name (str): name of collection.

        Returns:
            list[Stream]: list of streams in collection.
        """
        streams = self.reader_c.list_streams(collection_name)
        return streams

    def replay_collection_stream(
        self,
        collection_name: str,
        stream_name: str,
        attrs: dict,
        sink: Queue,
        transform=None,
    ) -> StreamAck:
        """
        Replay a collection-stream into a given queue.

        Args:
            collection_name (str): name of collection.
            stream_name (str): name of stream in collection.
            attrs (dict): attributes specifying which recording to replay.
            sink (asyncio.Queue): destination to buffer replayed data.
            transform (Callable): optional transform to apply to each data point.

        Returns:
            StreamAck: status and reference information.
        """
        randseq = prefix + gen_randseq()
        t = (lambda x: [randseq.encode(), *transform(x)]) if transform is not None else None
        self.context[randseq] = Event()
        self.reader_c.replay(collection_name, stream_name, attrs, sink, t, self.context[randseq])
        return StreamAck(status=True, randseq=randseq)

    def stop_task(
        self,
        randseq: str,
    ) -> StreamAck:
        if randseq in self.context:
            flag = self.context.pop(randseq)
            flag.set()
            return StreamAck(status=True)
        return StreamAck(status=True)

    def publish_collection_stream(
        self,
        collection_name: str,
        stream_name: str,
        attrs: dict,
    ) -> StreamAck:
        """
        Publish a collection-stream as a LSL stream.

        Args:
            collection_name (str): name of collection.
            stream_name (str): name of stream in collection.
            attrs (dict): attributes specifying which recording to publish.

        Returns:
            StreamAck: status and reference information.
        """
        self.reader_c.restream(collection_name, stream_name, attrs)
        return StreamAck(status=True)

    def list_live_streams(
        self,
    ) -> list[Stream]:
        """
        List all live streams.

        Returns:
            list[Stream]: list of available live streams.
        """
        self.reader_n.refresh_streams()
        streams = self.reader_n.list_streams()
        return streams

    def read_live_stream(
        self,
        stream_name: str,
        attrs: dict,
        sink: Queue,
        transform=None,
    ) -> StreamAck:
        """
        Read data from a live stream (LSL) into a given queue.

        Args:
            stream_name (str): name of live stream.
            attrs (dict): attributes specifying which live stream to read.
            sink (asyncio.Queue): destination to buffer replayed data.
            transform (Callable): optional transform to apply to each data point.

        Returns:
            StreamAck: status and reference information.
        """
        randseq = prefix + gen_randseq()
        t = (lambda x: [randseq.encode(), *transform(x)]) if transform is not None else None
        self.context[randseq] = Event()
        self.reader_n.relay(stream_name, attrs, sink, t, self.context[randseq])
        return StreamAck(status=True, randseq=randseq)



import asyncio
from abc import ABC, abstractmethod

class RpcClient(ABC):
    @abstractmethod
    async def connect(self, server_host: str, server_port: int) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def disconnect(self) -> None:
        raise NotImplementedError()


def create_rpc_client(
    name: str,
    codec_name: str,
    incoming: asyncio.Queue,
    outgoing: asyncio.Queue,
) -> RpcClient:
    assert name == "websocket"
    from .websocket import WebsocketRPC

    return WebsocketRPC(
        codec_name=codec_name,
        incoming=incoming,
        outgoing=outgoing,
    )



#!/usr/bin/env python3

import asyncio
import logging
import multiprocessing

import datamux.util as util
from datamux.api import DataMuxAPI
from rich.logging import RichHandler


async def main():
    N = 10
    P = 8000
    api = DataMuxAPI()

    # test 1 - collections
    logger.info("listing collections")
    collections = api.list_collections()
    logger.info(f"received collections: {collections}")

    # test 2 - collection streams
    logger.info("listing collection streams - battleship_gaze")
    collection_streams = api.list_collection_streams("battleship_gaze")
    logger.info(f"received collection_streams: {collection_streams}")

    # test 3 - replay collection stream
    logger.info("replaying collection stream")
    collection_name = "battleship_gaze"
    stream_name = "Gaze"
    attrs = dict(
        {
            "session": "1",
            "stream_name": "gaze",
        }
    )
    sink = multiprocessing.Queue()
    ack = api.replay_collection_stream(collection_name, stream_name, attrs, sink)
    assert ack.randseq is not None
    logger.info(f"received ack for collection stream: {ack}")
    # print P points or until EOF
    for _ in range(P):
        item = sink.get()
        logger.info(item)
        if item == util.END_OF_STREAM:
            break
    ack = api.stop_task(ack.randseq)

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(message)s", datefmt="[%X]", handlers=[RichHandler()])
    logger = logging.getLogger(__name__)
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        logger.warning("Interrupt received, shutting down.")



TOPIC_LIST_COLLECTIONS: bytes = b"lc"
TOPIC_LIST_COLLECTION_STREAMS: bytes = b"lcs"
TOPIC_REPLAY_COLLECTION_STREAM: bytes = b"rcs"
TOPIC_PUBLISH_COLLECTION_STREAM: bytes = b"pcs"
TOPIC_LIST_LIVE_STREAMS: bytes = b"lls"
TOPIC_READ_LIVE_STREAM: bytes = b"rls"
TOPIC_STOP_TASK: bytes = b"st"



from asyncio import Server


async def start_server(name: str) -> Server:
    if name == "empatica_e4":
        from .empatica_e4ss import start_server

        srv = await start_server()
        return srv
    else:
        raise ValueError(name)



import argparse
import asyncio
import logging

from rich.logging import RichHandler

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG, format="%(message)s", datefmt="[%X]", handlers=[RichHandler()])
    logger = logging.getLogger(__name__)

    parser = argparse.ArgumentParser(prog="DataMux", description="CLI for StreamingHub DataMux")

    parser.add_argument("command", choices=["serve", "init", "replay"])
    parser.add_argument("-H", "--host", type=str)
    parser.add_argument("-p", "--port", type=int)
    parser.add_argument("-r", "--rpc", type=str, choices=["websocket"])
    parser.add_argument("-c", "--codec", type=str, choices=["avro", "json"])
    parser.add_argument("--data_dir", type=str)
    parser.add_argument("--meta_dir", type=str)
    
    args = parser.parse_args()

    # Serving Remote API
    if args.command == "serve":
        from .serve import serve

        try:
            assert args.host is not None
            assert args.port is not None
            assert args.rpc is not None
            assert args.codec is not None
            asyncio.run(serve(args.host, args.port, args.rpc, args.codec))
        except KeyboardInterrupt:
            logging.warning("Interrupt received, shutting down.")

    # Initializing .streaminghubrc
    if args.command == "init":
        import json
        from pathlib import Path

        assert args.data_dir is not None
        assert args.meta_dir is not None
        config = dict(data_dir=args.data_dir, meta_dir=args.meta_dir)
        fp = Path("~/.streaminghubrc").expanduser().resolve()
        with open(fp, "w") as f:
            json.dump(config, f)
        logging.info("Updated ~/.streaminghubrc")



#!/usr/bin/env python3

"""
This command-line executable program accepts a
meta-stream as an input, and generates a
data-stream based on the given meta-stream spec.

It can be used for testing on meta-streams and
data-streams without the need for an actual
sensory device.
"""

import asyncio
import logging
import random
import sys
import time

import dfds
from rich.logging import RichHandler

SYNTAX = "datasource_simulate [path/to/datasource/spec]"

logger = logging.getLogger(__name__)


async def emit(source_id: str, spec: dfds.DataSourceSpec, stream_id: str):
    stream = spec.streams[stream_id]
    f = stream.frequency
    outlet = dfds.create_outlet_for_stream(source_id, spec.device, stream_id, stream)
    logger.info(f"created stream: {stream.name}")
    while True:
        # calculate dt from f. if f=0, assign a random dt
        dt = (1.0 / f) if f > 0 else (random.randrange(0, 10) / 10.0)
        if outlet.have_consumers():
            t1 = time.time_ns() * 1e-9
            sample = [random.gauss(0, random.random() / 2) for _ in range(len(stream.channels))]
            logger.debug(f"DataSource [{source_id}]: T={t1}, Sample={sample}")
            outlet.push_sample(sample, t1)
            # offset dt by cpu time
            t2 = time.time_ns() * 1e-9
            dt -= t2 - t1
        await asyncio.sleep(dt)


async def begin_streaming_random_data(spec: dfds.DataSourceSpec):
    source_id = dfds.util.gen_random_source_id()
    try:
        logger.info(
            f"DataSource [{source_id}]: Device: {spec.device.model}, {spec.device.manufacturer} ({spec.device.category})"
        )
        # create a job for each stream defined in the meta-stream
        jobs = [emit(source_id, spec, stream_id) for stream_id in spec.streams]
        # start all jobs
        logger.info(f"DataSource [{source_id}]: Started streams")
        await asyncio.gather(*jobs)
    except KeyboardInterrupt:
        logger.info(f"DataSource [{source_id}]: Interrupt received. Closing streams...")
    finally:
        logger.info(f"DataSource [{source_id}]: Closed streams")


def main():
    # parse command-line args
    args = sys.argv
    assert len(args) == 2, f"Invalid Syntax.\nExpected: {SYNTAX}"
    spec_name = args[1]
    logger.info(f"DataSourceSpec: {spec_name}")
    # load DataSourceSpec
    logger.info("Loading DataSourceSpec...")
    spec = dfds.get_datasource_spec(spec_name)
    logger.info("Loaded")
    # start data stream
    asyncio.get_event_loop().run_until_complete(begin_streaming_random_data(spec))


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(message)s", datefmt="[%X]", handlers=[RichHandler()])
    try:
        main()
    except AssertionError as e:
        logger.error(f"Error: {e}", file=sys.stderr)



#!/usr/bin/env python3

import asyncio
import logging

from rich.logging import RichHandler

from datamux.remote.server import DataMuxServer


async def serve(
    host: str,
    port: int,
    rpc: str,
    codec: str,
):
    server = DataMuxServer(rpc_name=rpc, codec_name=codec)
    await server.start(host, port)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(message)s", datefmt="[%X]", handlers=[RichHandler()])
    logger = logging.getLogger(__name__)

    host = "localhost"
    port = 3300
    rpc = "websocket"
    codec = "avro"
    try:
        asyncio.run(serve(host, port, rpc, codec))
    except KeyboardInterrupt:
        logger.warning("Interrupt received, shutting down.")



import asyncio
import http
import logging
from collections import defaultdict
from typing import Optional, Tuple

from websockets.datastructures import Headers, HeadersLike
from websockets.exceptions import ConnectionClosed
from websockets.server import WebSocketServerProtocol, serve

from ..codec import Codec, create_codec
from . import RpcServer


class WebsocketRPC(RpcServer):
    """
    WebSocket RPC server

    """

    def __init__(
        self,
        codec_name: str,
        incoming: asyncio.Queue,
        outgoing: asyncio.Queue,
    ) -> None:
        """
        Args:
            codec_name (str) codec to encode/decode outgoing/incoming messages
            incoming (asyncio.Queue) incoming messages from ALL clients
            outgoing (asyncio.Queue) outgoing messages for ALL clients
        """
        self.active = False
        # multiplexed queue
        self.incoming = incoming
        self.outgoing = outgoing
        # demultiplexed queues
        self.demux: dict[bytes, asyncio.Queue] = defaultdict(lambda: asyncio.Queue())
        # codec module
        self.codec_name = codec_name
        self.logger = logging.getLogger(__name__)

    async def __handle_demux__(self):
        """
        Demultiplex global outgoing to specific sources

        """
        self.logger.info("started demultiplexer")
        while self.active:
            (topic, content, id) = await self.outgoing.get()
            try:
                await self.demux[id].put((topic, content))
            except:
                self.logger.error(f"error in outgoing for id={id}")

    async def __handle_outgoing__(
        self,
        websocket: WebSocketServerProtocol,
        codec: Codec,
    ):
        """
        Write message from queue into the websocket

        Args:
            websocket (WebSocketServerProtocol): websocket connection

        """
        id = websocket.id.bytes
        try:
            while self.active:
                (topic, content) = await self.demux[id].get()
                self.logger.debug(f">: {topic}: {content}")
                msg = codec.encode(topic, content)
                if isinstance(msg, bytes):
                    await websocket.send(msg)
                if isinstance(msg, list):
                    for frame in msg:
                        await websocket.send(frame)
        except ConnectionClosed:
            self.logger.info(f"client connection closed: {id}")

    async def __handle_incoming__(
        self,
        websocket: WebSocketServerProtocol,
        codec: Codec,
    ):
        """
        Read content from websocket into the queue

        Args:
            websocket (WebSocketServerProtocol): websocket connection

        """
        id = websocket.id.bytes
        try:
            while self.active:
                msg = await websocket.recv()
                assert isinstance(msg, bytes)
                if (msg := codec.decode(msg)) is not None:
                    topic, content = msg
                    self.logger.debug(f"<: {topic}: {content}")
                    await self.incoming.put((topic, content, id))
        except ConnectionClosed:
            self.logger.info(f"client connection closed: {id}")

    async def __handle__(
        self,
        websocket: WebSocketServerProtocol,
    ):
        id = websocket.id.bytes
        codec = create_codec(self.codec_name)
        self.logger.info(f"client connected: {websocket.id}")
        outgoing = asyncio.create_task(self.__handle_outgoing__(websocket, codec))
        incoming = asyncio.create_task(self.__handle_incoming__(websocket, codec))
        done, pending = await asyncio.wait([outgoing, incoming], return_when=asyncio.FIRST_COMPLETED)
        for task in pending:
            task.cancel()
        self.logger.info(f"client disconnected: {id}")

    async def __intercept__(
        self,
        path: str,
        headers: Headers,
    ) -> Optional[Tuple[http.HTTPStatus, HeadersLike, bytes]]:
        if path == "/ws":
            return None
        elif path == "/":
            return http.HTTPStatus(200), [], b""
        else:
            return http.HTTPStatus(404), [], b""

    async def start(
        self,
        host: str,
        port: int,
    ):
        self.active = True
        self.server = await serve(
            ws_handler=self.__handle__,
            host=host,
            port=port,
            process_request=self.__intercept__,
        )
        asyncio.create_task(self.__handle_demux__())

    async def stop(
        self,
    ):
        self.active = False
        self.server.close()



from .codec import Codec


def create_codec(
    backend: str,
) -> Codec:
    if backend == "json":
        from .json import JSONCodec

        return JSONCodec()
    if backend == "avro":
        from .avro import AvroCodec

        return AvroCodec()
    raise NotImplementedError()



#!/usr/bin/env python3

"""
Mock E4 Streaming Server
Source: git@github.com:nirdslab/ee4ss_pretender.git

@author: Ben Mabe, Yasith Jayawardana

"""

import asyncio
import logging
import sys
from asyncio.streams import StreamReader, StreamWriter
from datetime import datetime
from threading import Thread
from typing import Dict


class Stream:
    def __init__(self, label: str, typ: str, freq: float, default_v: list):
        self.label = label
        self.typ = typ
        self.freq = freq or 0.1
        self.value = default_v
        self.last_t = 0.0
        self.last_v = None

    def read_latest(self):
        t = datetime.now().timestamp()
        if (t - self.last_t) > 1 / self.freq:
            self.last_t = t
            self.last_v = " ".join(map(str, self.value))
        return f"{self.label} {self.last_t} {self.last_v}"


STREAM_TYPES = ["acc", "bvp", "gsr", "ibi", "tmp", "bat", "tag", "hr"]
STREAMS: Dict[str, Stream] = {
    "acc": Stream("E4_Acc", "acc", 32, [100, 200, 300]),
    "bvp": Stream("E4_Bvp", "bvp", 64, [400]),
    "ibi": Stream("E4_Ibi", "ibi", 0, [500]),
    "gsr": Stream("E4_Gsr", "gsr", 4, [600]),
    "tmp": Stream("E4_Temperature", "tmp", 4, [700]),
    "hr": Stream("E4_Hr", "hr", 0, [800]),
    "bat": Stream("E4_Bat", "bat", 0, [1.0]),
    "tag": Stream("E4_Tag", "tag", 0, [1000]),
}


class SubscriptionContext:
    def __init__(self, r: StreamReader, w: StreamWriter) -> None:
        self.r = r
        self.w = w
        self.keepalive = True
        self.subscriptions = []
        self.paused = False
        self.ctx_thread = Thread(target=self.create_streaming_loop)
        self.ctx_thread.start()

    def subscribe(self, name: str):
        if name not in self.subscriptions:
            self.subscriptions.append(name)
        else:
            logger.warning(f"attempt to resubscribe to {name}")

    def unsubscribe(self, name: str):
        if name in self.subscriptions:
            self.subscriptions.remove(name)
        else:
            logger.warning(f"attempt to unsubscribe from non-subscribed {name}")

    def end(self):
        self.keepalive = False
        self.ctx_thread.join()

    async def __task__(self, stream_id: str):
        s = STREAMS[stream_id]
        while self.keepalive:
            if not self.paused and s.typ in self.subscriptions:
                data = s.read_latest()
                self.w.write(f"{data}\r\n".encode())
            await asyncio.sleep(1 / s.freq)

    def create_streaming_loop(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        tasks = []
        for stream_id in STREAMS:
            tasks.append(loop.create_task(self.__task__(stream_id)))
        loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))


def cmd_handler(msg: str, ctx: SubscriptionContext) -> str:
    # If msg does not exist, or cmd empty, return empty
    if not msg:
        return ""

    msg_parts = msg.split()
    cmd, args = msg_parts[0], msg_parts[1:]

    if cmd == "device_list":
        return "R device_list 2 | 9ff167 Empatica_E4 | 7a3166 Empatica_E4"

    if cmd == "device_connect":
        return "R device_connect OK"

    if cmd == "device_subscribe":
        if len(args) > 2:
            return "R device_subscribe ERR too many arguments"
        sub = args[0].lower()
        if sub in STREAM_TYPES:
            ctx.subscribe(sub)
            return f"R {cmd} {sub} OK"

    if cmd == "device_unsubscribe":
        if len(args) > 2:
            return "R device_unsubscribe ERR too many arguments"
        sub = args[0].lower()
        if sub in STREAM_TYPES:
            ctx.unsubscribe(sub)
            return f"R {cmd} {sub} OK"

    if cmd == "pause":
        if len(args) > 1:
            return "R pause ERR too many arguments"
        else:
            if args[0].upper() == "ON":
                if ctx.paused:
                    return "R pause ERR already paused"
                ctx.paused = True
                return "R pause ON"
            elif args[0].upper() == "OFF":
                if not ctx.paused:
                    return "R pause ERR not paused"
                ctx.paused = False
                return "R pause OFF"
            else:
                return "R pause ERR wrong argument"

    return ""


async def handler(r: StreamReader, w: StreamWriter):
    logger.info(f"connection established")
    # create new context for client
    ctx = SubscriptionContext(r, w)
    try:
        while not r.exception():
            msg_raw = await r.readline()
            msg = msg_raw.decode().strip()
            logger.info(f"< {msg}")
            out = cmd_handler(msg, ctx)
            logger.info(f"> {out}")
            w.write(f"{out}\r\n".encode())
            await w.drain()
    except UnicodeDecodeError:
        logger.warning("invalid char encountered")
    except AttributeError:
        logger.warning("attribute error encountered")
    except ConnectionError:
        logger.warning("connection error encountered")
    finally:
        logger.info("cleaning up resources")
        ctx.end()
    logger.info(f"resources purged. connection terminated.")


async def start_server() -> asyncio.Server:
    srv = await asyncio.start_server(handler, "localhost", 28000)
    return srv


if __name__ == "__main__":
    logger = logging.getLogger(__name__)
    logging.basicConfig(level=logging.DEBUG, format="%(message)s", datefmt="[%X]", handlers=[logging.StreamHandler(sys.stdout)])

    async def main():
        srv = await start_server()
        host, port = srv.sockets[0].getsockname()
        logger.info(f"Streaming server started on {host}:{port}!")
        try:
            async with srv:
                await srv.serve_forever()
        except KeyboardInterrupt:
            await srv.wait_closed()
            exit(0)

    asyncio.run(main(), debug=True)



import logging
from multiprocessing import Queue
from threading import Event, Thread

import pylsl
from dfds.typing import Stream

import datamux.util as util

from . import Reader
from .util import stream_info_to_stream, stream_inlet_to_stream


class NodeReader(Reader):
    """
    Stream Reader for DFDS Nodes

    Functions Provided:

    * refresh_streams()

    * list_streams()

    * relay(stream_id, queue)

    """

    __streams: list[Stream] = []
    logger = logging.getLogger(__name__)

    def refresh_streams(
        self,
    ) -> None:
        self.__streams.clear()
        stream_infos = pylsl.resolve_streams()
        for stream_info in stream_infos:
            stream = stream_info_to_stream(stream_info)
            self.__streams.append(stream)

    def list_streams(
        self,
    ) -> list[Stream]:
        return self.__streams

    def relay(
        self,
        stream_name: str,
        attrs: dict,
        queue: Queue,
        transform,
        flag,
    ) -> None:
        stream = None
        for s in self.__streams:
            is_name_matched = s.name == stream_name
            is_attrs_matched = dict(s.attrs) == attrs
            if is_name_matched and is_attrs_matched:
                stream = s
        assert stream
        query = self.__create_query(stream)
        self.logger.info("query: %s", query)

        # get the LSL stream for the DFDS stream
        stream_infos = pylsl.resolve_bypred(query)
        self.logger.info("found %d stream(s)", len(stream_infos))
        assert len(stream_infos) == 1
        stream_info = stream_infos[0]

        # create task to live-stream data
        self.logger.info("creating inlet")
        inlet = pylsl.StreamInlet(stream_info)
        self.logger.info("creating relay task")
        thread = Thread(None, self.relay_coro, stream_name, (inlet, queue, transform, flag), daemon=True)
        thread.start()

    def __create_query(
        self,
        stream: Stream,
    ) -> str:
        query_args = {}
        query_args["name"] = stream.name
        for k, v in stream.attrs.items():
            query_args[f"desc/attrs/{k}"] = v
        query_str = " and ".join([f"{k}='{v}'" for k, v in query_args.items()])
        return query_str

    def relay_coro(
        self,
        inlet: pylsl.StreamInlet,
        queue: Queue,
        transform,
        flag: Event,
    ):
        stream = stream_inlet_to_stream(inlet)
        if not "dfds_mode" in stream.attrs:
            stream.attrs.update(dfds_mode="relay")

        # column names
        index_cols = list(stream.index)
        value_cols = list(stream.fields)

        # termination indicator
        eof = util.END_OF_STREAM
        if transform is not None:
            eof = transform(eof)

        # relay each record
        self.logger.info("relay started")
        while True:
            if flag.is_set():
                self.logger.info(f"relay stop requested")
                break

            try:
                (values, indices) = inlet.pull_chunk()
            except pylsl.LostError as e:
                self.logger.info(f"LSL connection lost: {e}")
                break
            if values is None or len(values) == 0:
                pass
            else:
                for index, value in zip(indices, values):
                    # FIXME currently only supports 1D index
                    index_dict = {index_cols[0]: index}
                    value_dict = dict(zip(value_cols, value))
                    msg = dict(index=index_dict, value=value_dict)
                    if transform is not None:
                        msg = transform(msg)
                    queue.put_nowait(msg)

        inlet.close_stream()
        queue.put_nowait(eof)
        self.logger.info("relay ended")
